from typing import Dict, Type, Tuple, Optional, Callable

from autograder_platform.Executors.Environment import ExecutionEnvironment, ImplEnvironment, ImplResults

from autograder_platform.StudentSubmission.ISubmissionProcess import ISubmissionProcess

from autograder_platform.StudentSubmission.AbstractStudentSubmission import AbstractStudentSubmission
from autograder_platform.Tasks.TaskRunner import TaskRunner
from autograder_platform.config.Config import AutograderConfiguration


class SubmissionProcessFactory:
    registry: \
        Dict[Type[AbstractStudentSubmission],
        Tuple[Type[ISubmissionProcess],
        Optional[Type[ImplEnvironment]],
        Optional[Callable[[ImplEnvironment, AutograderConfiguration], None]]]] \
        = {}

    @classmethod
    def register(cls, submission: Type[AbstractStudentSubmission], process: Type[ISubmissionProcess],
                 implEnvironment: Optional[Type[ImplEnvironment]] = None,
                 implEnvironmentConfigMapper: Optional[
                     Callable[[ImplEnvironment, AutograderConfiguration], None]] = None) -> None:

        if submission in cls.registry:
            return

        if not issubclass(process, ISubmissionProcess):
            raise AttributeError(f"{process} is not a subclass of ISubmissionProcess! Registration failed!")

        if not issubclass(submission, AbstractStudentSubmission):
            raise AttributeError(f"{submission} is not a subclass of AbstractStudentSubmission! Registration failed!")

        if implEnvironment is not None and implEnvironmentConfigMapper is None:
            raise AttributeError(
                f"Implementation environment is provided for submission type {submission}, but no mapper is defined! Registration Failed!")

        if implEnvironment is None and implEnvironmentConfigMapper is not None:
            raise AttributeError(
                f"Implementation environment mapper is provided for submission type {submission}, but no implementation environment is defined! Registration Failed!")

        # TODO Add logging here
        cls.registry[submission] = (process, implEnvironment, implEnvironmentConfigMapper)

    @classmethod
    def createProcess(cls, environment: ExecutionEnvironment[ImplEnvironment, ImplResults],
                      runner: TaskRunner, autograderConfig: AutograderConfiguration) -> ISubmissionProcess:
        submissionType = runner.getSubmissionType()

        if submissionType not in cls.registry.keys():
            raise AttributeError(f"{submissionType} has not been registered. Lookup failed.")

        processType, implEnvironmentType, mapper = cls.registry[submissionType]

        if implEnvironmentType is not None and mapper is not None:
            if environment.impl_environment is None:
                environment.impl_environment = implEnvironmentType()

            mapper(environment.impl_environment, autograderConfig)

        process = processType()

        process.setup(environment, runner)

        return process



from collections import deque
from typing import Dict, List, Optional, Type, Deque

from autograder_platform.StudentSubmission.AbstractStudentSubmission import AbstractStudentSubmission
from autograder_platform.Tasks.Task import Task
from autograder_platform.Tasks.common import TaskAlreadyExists, TaskDoesNotExist, TaskStatus


class TaskRunner:
    def __init__(self, submissionType: Type[AbstractStudentSubmission]):
        self.tasks: Dict[str, Task] = {}
        self.order: Deque[str] = deque()
        self.overallResultTask: Optional[str] = None
        self.errorOccurred = False
        self.submissionType: Type[AbstractStudentSubmission] = submissionType

    def add(self, task: Task, isOverallResultTask: bool = False):
        if task.getName() in self.tasks:
            raise TaskAlreadyExists(task.getName())

        self.tasks[task.getName()] = task
        self.order.append(task.getName())

        if isOverallResultTask:
            self.overallResultTask = task.getName()

    def getResult(self, taskName: str) -> object:
        if taskName not in self.tasks:
            raise TaskDoesNotExist(taskName)

        return self.tasks[taskName].getResult()


    def run(self) -> object:
        result: object = None
        while self.order:
            task: Task = self.tasks[self.order.popleft()]

            task.doTask()

            if task.getStatus() != TaskStatus.COMPLETE:
                self.errorOccurred = True
                break

            if task.getName() == self.overallResultTask:
                result = task.getResult()

        return result

    def wasSuccessful(self):
        return not self.order and not self.errorOccurred

    def getAllErrors(self) -> List[Exception]:
        errors: List[Exception] = []
        for task in self.tasks.values():
            error = task.getError()
            if error is None:
                continue

            errors.append(error)


        return errors

    def getSubmissionType(self) -> Type[AbstractStudentSubmission]:
        return self.submissionType

    def __getstate__(self):
        return {"tasks": self.tasks, "order": self.order, "overallResultTask": self.overallResultTask, "errorOccurred": self.errorOccurred}

    def __setstate__(self, state):
        self.tasks = state["tasks"]
        self.order = state["order"]
        self.overallResultTask = state["overallResultTask"]
        self.errorOccurred = state["errorOccurred"]



import hashlib
import json
import os
import re
import subprocess
import sys
from typing import Dict, List, Optional

import BetterPyUnitFormat
import tomli

from autograder_platform.cli import AutograderCLITool
from autograder_platform.config.Config import AutograderConfigurationBuilder, AutograderConfiguration, \
    AutograderConfigurationProvider


class LocalAutograderCLI(AutograderCLITool):
    SUBMISSION_REGEX: re.Pattern = re.compile(r"^(\w|\s)+\.py$")
    FILE_HASHES_NAME = ".filehashes"

    def __init__(self):
        super().__init__(f"Local v{AutograderCLITool.get_version()}")

        self.config_location = ""

    def clean_previous_submissions(self, directory: str):
        """
        This function cleans out previous submissions if they exist.
        :param directory: the directory to run the detection in.
        """
        zip_files = [os.path.join(directory, file) for file in os.listdir(directory) if file[-4:] == ".zip"]
        if len(zip_files) > 0:
            self.print_info_message("Previous submissions found. Cleaning out old submission files")
            for file in zip_files:
                os.remove(file)

    def generate_hashes(self, submission_directory: str):
        """
        This function generates hashes for the files in submission directory.

        :param submission_directory: The directory to base all the hashes out of
        """

        pythonFiles = [
            os.path.join(submission_directory, file)
            for file in os.listdir(submission_directory)
            if self.SUBMISSION_REGEX.match(file)
        ]

        fileHashes: dict[str, str] = {}

        for file in pythonFiles:
            with open(file, 'rb') as r:
                # technically, this can lead to overflow issues if the file is too large.
                #  That shouldn't happen in this class hopefully
                fileBytes = r.read()
                fileHashes[file] = hashlib.md5(fileBytes, usedforsecurity=False).hexdigest()

        return fileHashes

    def verify_file_changed(self, submission_directory: str):
        """
        This function checks to see if a file has changed since the last run of this script.

        It stores its results in `submissionDirectory/.filehashes`. Currently it does not support sub directories.

        :param submission_directory: the directory that the student is doing their work in.

        :return: true if at least of the files changed.

        """

        FILE_HASHES_PATH = os.path.join(submission_directory, self.FILE_HASHES_NAME)

        if not os.path.exists(FILE_HASHES_PATH):
            with open(FILE_HASHES_PATH, 'w') as w:
                json.dump({}, w)

        with open(FILE_HASHES_PATH, 'r') as r:
            try:
                existingHashes = json.load(r)
            except json.JSONDecodeError:
                existingHashes = None

        newHashes = self.generate_hashes(submission_directory)

        with open(FILE_HASHES_PATH, 'w') as w:
            json.dump(newHashes, w)

        return existingHashes != newHashes

    def verify_student_work_present(self, submission_directory: str) -> bool:
        if not os.path.exists(submission_directory):
            self.print_error_message(self.SUBMISSION_ERROR, f"Failed to locate student work in {submission_directory}")
            return False

        # this doesn't catch files in folders. Something to be aware of for students
        files = [file for file in os.listdir(submission_directory) if self.SUBMISSION_REGEX.match(file)]

        if len(files) < 1:
            self.print_error_message(self.SUBMISSION_ERROR, f"No valid files found in submission directory.")
            self.print_error_message(self.SUBMISSION_ERROR, f"Found {os.listdir(submission_directory)}.")
            return False

        return True

    @staticmethod
    def discover_autograders(current_root: str, autograders: List[str]):
        contents = os.listdir(current_root)

        for path in contents:
            path = os.path.join(current_root, path)

            if "__" in path:
                continue

            if os.path.basename(path)[0] == ".":
                continue

            if os.path.isfile(path) and os.path.basename(path) == "config.toml":
                autograders.append(path)
                break

            if os.path.isdir(path):
                LocalAutograderCLI.discover_autograders(path, autograders)

    @staticmethod
    def get_autograder_name(path) -> Optional[str]:
        with open(path, 'rb') as rb:
            data = tomli.load(rb)

        if "assignment_name" not in data or not data["assignment_name"]:
            return None

        return data["assignment_name"]

    @staticmethod
    def get_autograder_version(path) -> Optional[str]:
        with open(path, 'rb') as rb:
            data = tomli.load(rb)

        if "config" not in data or "autograder_version" not in data["config"]:
            return None

        return data["config"]["autograder_version"]

    def compare_autograder_versions(self, required_version: str) -> bool:
        version = list(map(int, required_version.split(".")))
        actual_version = list(map(int, self.get_version().split(".")))

        status = False
        if version[0] != actual_version[0]:
            return False

        if actual_version[1] > version[1]:
            status = True
        elif actual_version[1] == version[1] and actual_version[2] >= version[2]:
            status = True

        return status

    def update_autograder(self, version) -> bool:
        self.print_info_message("Updating autograder...")

        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", f"128Autograder=={version}", "--break-system-packages"])
        except subprocess.CalledProcessError:
            self.print_error_message(self.ENVIRONMENT_ERROR, "Failed to update autograder!")
            return False

        self.print_info_message("Autograder Updated")
        return True

    def select_root(self) -> Optional[str]:
        autograders = []

        full_path = os.path.abspath(".")

        self.print_info_message(f"Discovering autograders in {full_path}...")

        self.discover_autograders(full_path, autograders)

        if len(autograders) == 0:
            self.print_error_message(self.ENVIRONMENT_ERROR, f"Failed to locate any autograders in '{full_path}'!")
            self.print_error_message(self.ENVIRONMENT_ERROR,
                                     "Make sure that this script is being run in the same directory as your autograder!")
            return None

        if len(autograders) == 1:
            self.print_info_message(f"Selecting autograder config located at '{autograders[0]}'")
            return autograders[0]

        autograders.sort()

        self.print_info_message(f"Multiple autograders found at {full_path}!")
        self.print_info_message(f"Please select the autograder you want to run")

        for i, path in enumerate(autograders):
            name = self.get_autograder_name(path)

            if name is None:
                continue

            print(f"[{i + 1:02}] {name}")

        selection = len(autograders) + 1

        while not (1 <= selection <= len(autograders)):
            try:
                selection = int(
                    input(f"Enter number corresponding to the autograder to run (1 - {len(autograders)}): "))
            except ValueError:
                selection = len(autograders) + 1

        self.print_info_message(f"Selecting autograder config located at '{autograders[selection - 1]}'")
        return autograders[selection - 1]

    def configure_options(self):  # pragma: no cover
        self.parser.add_argument("--submission-directory", default="student_work",
                                 help="The location for the student's submission relative to the submission root")
        self.parser.add_argument("--test-directory", default="student_tests",
                                 help="The location for the tests for the autograder relative to the submission root")
        self.parser.add_argument("--bypass-version-check", action="store_true", default=False,
                                 help="Bypass autograder version verification. Note: This may cause the autograder to fail!")
        self.parser.add_argument("--version", action="store_true", default=False, help="Print out version and exit")

    def set_config_arguments(self, configBuilder: AutograderConfigurationBuilder[AutograderConfiguration]):  # pragma: no cover
        pass

    def run(self) -> bool:  # pragma: no cover
        self.configure_options()

        self.arguments = self.parser.parse_args()

        if self.arguments.version:
            self.print_info_message(f"Autograder version: {self.get_version()}")
            return False

        self.config_location = os.path.abspath(self.arguments.config_file) if \
            os.path.exists(self.arguments.config_file) else self.select_root()

        if self.config_location is None:
            self.print_error_message(self.ENVIRONMENT_ERROR, "Failed to load autograder!")
            return True

        root_directory = os.path.dirname(self.config_location)

        self.print_info_message(f"Running autograder from '{root_directory}'")

        self.print_info_message("Verifying autograder version")
        version = self.get_autograder_version(self.config_location)

        if not self.arguments.bypass_version_check and not self.compare_autograder_versions(version):
            if self.update_autograder(version):
                self.print_info_message("Updated succeeded! Please rerun the script!")
                return True
            else:
                self.print_error_message(self.ENVIRONMENT_ERROR,
                                         "Update failed! Please see above for failure reason or rerun as 'test_my_work --bypass-version-check'")
                return True

        if not self.verify_student_work_present(os.path.join(root_directory, self.arguments.submission_directory)):
            return True

        fileChanged = self.verify_file_changed(os.path.join(root_directory, self.arguments.submission_directory))

        self.config = AutograderConfigurationBuilder() \
            .fromTOML(self.config_location) \
            .setAutograderRoot(root_directory) \
            .setStudentSubmissionDirectory(os.path.join(root_directory, self.arguments.submission_directory)) \
            .setTestDirectory(os.path.join(root_directory, self.arguments.test_directory)) \
            .build()

        AutograderConfigurationProvider.set(self.config)

        self.discover_tests()

        self.print_info_message("Starting autograder")

        runner = BetterPyUnitFormat.BetterPyUnitTestRunner()

        res = runner.run(self.tests)

        if not fileChanged:
            self.print_warning_message("Student Submission Warning", "Student's submission may not have changed!")

        return not res.wasSuccessful()


tool = LocalAutograderCLI().run

if __name__ == "__main__":
    res = tool()

    # this is the same behavior as the script and isn't actually documented well!
    # thanks python
    exit(res)




import dataclasses
from typing import List, Dict, Optional, TypeVar
from importlib import import_module

from autograder_platform.StudentSubmissionImpl.Python.AbstractPythonImportFactory import AbstractModuleFinder
from autograder_platform.StudentSubmissionImpl.Python.PythonModuleMockImportFactory import MockedModuleFinder
from autograder_platform.TestingFramework.SingleFunctionMock import SingleFunctionMock
from autograder_platform.config.Config import AutograderConfiguration


class PythonResults():
    class Mocks():
        def __init__(self, mocks: Optional[Dict[str, SingleFunctionMock]]):
            self.mocks = mocks

        def __getitem__(self, mockName: str) -> SingleFunctionMock:
            if self.mocks is None:
                raise AssertionError("No mocks were returned by student submission!")
            if mockName not in self.mocks:
                raise AssertionError(f"Mock '{mockName}' was not returned by the students submission")

            return self.mocks[mockName]

    def __init__(self, mocks=None):
        self.mocks = mocks

    @property
    def mocks(self) -> Mocks:
        return self._mocks

    @mocks.setter
    def mocks(self, value: Optional[Dict[str, SingleFunctionMock]]):
        self._mocks = PythonResults.Mocks(value)


@dataclasses.dataclass
class PythonEnvironment():
    buffer_size: int = 2**20
    """Buffer size set from the config file. This shouldn't be set directly"""
    import_loader: List[AbstractModuleFinder] = dataclasses.field(default_factory=list)
    """The import loader. This shouldn't be set directly"""
    mocks: Dict[str, Optional[SingleFunctionMock]] = dataclasses.field(default_factory=dict)
    """What mocks have been defined for this run of the student's submission"""


def configMapper(env: PythonEnvironment, config: AutograderConfiguration):
    if config.config.python is None:
        raise AttributeError("INVALID STATE: Implementation environment mapping FAILED! Python config is NONE when should be defined!")

    env.buffer_size = config.config.python.buffer_size


Builder = TypeVar("Builder", bound="PythonEnvironmentBuilder")


class PythonEnvironmentBuilder():
    def __init__(self) -> None:
        self.environment: PythonEnvironment = PythonEnvironment()
        self.moduleMocks: Dict[str, Dict[str, SingleFunctionMock]] = {}

    def addModuleMock(self: Builder, moduleName: str, mockedMethods: Dict[str, SingleFunctionMock]) -> Builder:
        """
        Description
        ---
        This function sets up a mock for a complete module. 
        All mocks must be the same 'level' meaning we cant mock a.b.fun and a.fun. We have to choose. 

        We also  cant mock both a.b and a in the same submission currently without mocking the entirety of a.

        :param moduleName: The name of the module that will be mocked.
        :param mockedMethods: the map of the methods to mock in the module
        """
        if moduleName in self.moduleMocks:
            for mockName, mockObject in mockedMethods.items():
                self.moduleMocks[moduleName][mockName] = mockObject

            return self

        self.moduleMocks[moduleName] = mockedMethods

        return self

    def addImportHandler(self: Builder, importHandler: AbstractModuleFinder) -> Builder:
        """
        Description
        ---
        This adds an import handler to the environment

        :param importHandler: the meta path finder
        """
        self.environment.import_loader.append(importHandler)

        return self

    def addMock(self: Builder, mockName: str, mockObject: SingleFunctionMock) -> Builder:
        """
        This needs to be updated once we decide how to do mocks
        """
        self.environment.mocks[mockName] = mockObject

        return self

    def _processAndValidateModuleMocks(self):
        for moduleName in self.moduleMocks.keys():
            try:
                module = import_module(moduleName)
            except ImportError:
                raise AttributeError(f"Failed to import {moduleName}!")

            for methodName, mock in self.moduleMocks[moduleName].items():

                if not isinstance(mock, SingleFunctionMock):
                    raise AttributeError(f"Invalid mock for {methodName}")

                self.environment.mocks[methodName] = None

            self.environment.import_loader.append(MockedModuleFinder(moduleName, module, self.moduleMocks[moduleName]))

    def build(self) -> PythonEnvironment:
        self._processAndValidateModuleMocks()

        return self.environment



import os

from autograder_utils.ResultBuilders import prairieLearnResultBuilder
from autograder_utils.ResultFinalizers import prairieLearnResultFinalizer
from autograder_utils.JSONTestRunner import JSONTestRunner

# CLI tools should only be able to import from the CLI part of the library
from autograder_platform.cli import AutograderCLITool
from autograder_platform.config.Config import AutograderConfigurationBuilder, AutograderConfiguration


class PrairieLearnAutograderCLI(AutograderCLITool):
    def __init__(self):
        super().__init__("PrairieLearn")

    def configure_options(self):  # pragma: no cover
        self.parser.add_argument("--results-location", default="/grade/results/results.json",
                                 help="The location for the autograder JSON results")
        self.parser.add_argument("--metadata-path", default="/grade/data/data.json",
                                 help="The location for the submission metadata JSON")
        self.parser.add_argument("--autograder-root", default="/grade/tests",
                                 help="The base location for the autograder")
        self.parser.add_argument("--test-directory", default="student_tests",
                                 help="The location for the student tests")
        self.parser.add_argument("--submission-directory", default="/grade/student",
                                 help="The directory where the student's submission is located")

    def set_config_arguments(self, configBuilder: AutograderConfigurationBuilder[AutograderConfiguration]):  # pragma: no cover
        if self.arguments is None:
            return

        configBuilder.setStudentSubmissionDirectory(self.arguments.submission_directory)
        configBuilder.setAutograderRoot(self.arguments.autograder_root)
        configBuilder.setTestDirectory(str(os.path.join(self.arguments.autograder_root, self.arguments.test_directory)))

    def run(self) -> bool:  # pragma: no cover
        self.configure_options()

        self.load_config()

        if self.arguments is None:
            return True

        self.discover_tests()

        with open(self.arguments.results_location, 'w') as w:
            testRunner = JSONTestRunner(visibility='visible', stream=w,
                                        result_builder=prairieLearnResultBuilder,
                                        result_finalizer=prairieLearnResultFinalizer)

            res = testRunner.run(self.tests)

            return not res.wasSuccessful()


tool = PrairieLearnAutograderCLI().run

if __name__ == "__main__":
    res = tool()

    exit(res)




import abc
from autograder_platform.Executors.Environment import ExecutionEnvironment

from autograder_platform.Tasks.TaskRunner import TaskRunner


class ISubmissionProcess(abc.ABC):
    @abc.abstractmethod
    def setup(self, environment: ExecutionEnvironment, runner: TaskRunner):
        pass

    @abc.abstractmethod
    def run(self):
        pass

    @abc.abstractmethod
    def populateResults(self, environment: ExecutionEnvironment):
        pass

    @abc.abstractmethod
    def cleanup(self):
        pass

    @classmethod
    @abc.abstractmethod
    def processAndRaiseExceptions(cls, environment: ExecutionEnvironment):
        pass



import abc
from typing import Generic, List, Set, TypeVar, Dict

from autograder_platform.StudentSubmission.common import ValidationError, ValidationHook

from autograder_platform.StudentSubmission.AbstractValidator import AbstractValidator
from autograder_platform.StudentSubmission.GenericValidators import SubmissionPathValidator

T = TypeVar("T")

# for some reason this has to be TBuilder??
TBuilder = TypeVar("TBuilder", bound="AbstractStudentSubmission[Any]")


class AbstractStudentSubmission(abc.ABC, Generic[T]):
    """
    Description
    ===========

    This class contains the abstract student submission.
    Basically, this enables more of a plug and play architechture for different submission models.

    This model also allows a cleaner and more consistent way to implement validation of the submission.

    You can implement :ref:`AbstractValidator` for your purpose and assign it a hook.
    Then when we reach that phase of the submission, that hook will be validated.

    Subclasses must implement ``doLoad`` and ``doBuild``.
    Subclasses must also implement ``getExecutableSubmission`` which should return the submission in a state that can be executed 
    by whatever runner is implemented.

    For an example, take a look at the Python (or coming soon, the C / C++) implementation.
    """
    def __init__(self):
        self.submissionRoot: str = "."
        self.validators: Dict[ValidationHook, Set[AbstractValidator]] = {}
        self.validationErrors: List[Exception] = []

        # default validators
        self.addValidator(SubmissionPathValidator())


    def setSubmissionRoot(self: TBuilder, submissionRoot: str) -> TBuilder:
        """
        Description
        ---
        
        Defines the root of the submission. 

        Should be a path to the ``student_work``

        :param submissionRoot: The root of the submission.
        :returns: self
        """
        self.submissionRoot = submissionRoot
        return self

    def addValidator(self: TBuilder, validator: AbstractValidator) -> TBuilder:
        """
        Description
        ---

        Adds a validator to run at the hook defined in the validator.

        Validators are provided a reference to the submission at setup time to gather 
        information that they need to know about the student's submission.
        
        The hook to use is determined by the abstract static method ``AbstractValidator.getValidationHook()``.

        Only one validator of each type is allowed, ie: If we pass two validators of type ValidateAST, 
        only the last one added will be run. This enforces single responsibity.

        :param validator: the validator to add subject to the above information.
        :returns: self
        """
        hook = validator.getValidationHook()

        if hook not in self.validators.keys():
            self.validators[hook] = set()

        self.validators[hook].add(validator)
        return self

    def _validate(self, validationHook: ValidationHook):
        if validationHook not in self.validators.keys():
            return

        for validator in self.validators[validationHook]:
            validator.setup(self)
            validator.run()
            self.validationErrors.extend(validator.collectErrors())

        if self.validationErrors:
            raise ValidationError(self.validationErrors)


    @abc.abstractmethod
    def doLoad(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def doBuild(self):
        raise NotImplementedError()
    
    def load(self: TBuilder) -> TBuilder:
        """
        Description
        ---

        Runs all validators attached to the ``LOAD`` hooks.

        This method calls the overriden ``doLoad`` method.

        This is step 1 in the submission pipeline.
        With the python version, this discovers submitted files and identifies the main module.

        The flow is:
        1. Run ``PRE_LOAD`` validation hook
        2. Run overriden ``doLoad``
        3. Run ``POST_LOAD`` validation hook

        :returns: self
        """
        self._validate(ValidationHook.PRE_LOAD)

        self.doLoad()

        self._validate(ValidationHook.POST_LOAD)
        return self

    def build(self: TBuilder) -> TBuilder:
        """
        Description
        ---
        
        Runs all validators attached to the ``BUILD`` hooks.

        This method calls the overriden ``doBuild`` method.

        This is step 2 in the submission pipeline.
        With the python version, this builds all the discovered files to the AST representation.

        The flow is:
        1. Run ``PRE_BUILD`` validation hook
        2. Run overriden ``doBuild``
        3. Run ``POST_BUILD`` validation hook

        :returns: self
        """
        self._validate(ValidationHook.PRE_BUILD)

        self.doBuild()

        self._validate(ValidationHook.POST_BUILD)
        return self

    def validate(self: TBuilder) -> TBuilder:
        """
        Description
        ---

        Runs all validators attached to the ``VALIDATION`` hook.

        This method should be overriden to perform any final validation that is needed prior to getting the executable version of the submission.

        :returns: self
        """
        self._validate(ValidationHook.VALIDATION)

        return self

    @abc.abstractmethod
    def getExecutableSubmission(self) -> T:
        pass

    def getSubmissionRoot(self) -> str:
        return self.submissionRoot




from importlib import import_module
from types import CodeType, ModuleType
from typing import TypeVar, Tuple, List, Final, Optional, Dict, Callable, TypedDict

from autograder_platform.StudentSubmission.common import InvalidRunner, MissingFunctionDefinition
from autograder_platform.StudentSubmissionImpl.Python import PythonSubmission
from autograder_platform.StudentSubmissionImpl.Python.common import PythonTaskResult
from autograder_platform.Tasks.TaskRunner import TaskRunner
from autograder_platform.Tasks.Task import Task
from autograder_platform.TestingFramework.SingleFunctionMock import SingleFunctionMock

Builder = TypeVar('Builder', bound="PythonRunnerBuilder")

class RunMethodResult(TypedDict):
    return_val: object
    parameters: Tuple[object, ...]


class Parameter:
    def __init__(self, value: Optional[object] = None, autowiredName: Optional[str] = None):
        if value is not None and autowiredName is not None:
            raise InvalidRunner("Value and Autowire cannot both be defined.")

        if value is None and autowiredName is None:
            raise InvalidRunner("Exactly one of value and autowire must be defined.")

        self._useAutowire: Final[bool] = autowiredName is not None

        self._autowire: Final[str] = autowiredName

        self._value: Final[object] = value

    @property
    def useAutowire(self) -> bool:
        return self._useAutowire

    @property
    def autowire(self) -> str:
        if not self._useAutowire:
            raise AttributeError("Attempt to access autowire when autowire is null!")
        return self._autowire

    @property
    def value(self) -> object:
        if self._useAutowire:
            raise AttributeError("Attempt to access value when value is null!")
        return self._value

    def get(self, module: ModuleType) -> object:
        if not self.useAutowire:
            return self.value

        autowiredParam: Optional[object] = getattr(module, self.autowire, None)

        if autowiredParam is None:
            raise RuntimeError(
                f"INVALID STATE: Failed to map '{self.autowire}' to type when resolving autowired parameters!")

        return autowiredParam


class PythonTaskLibrary:

    @staticmethod
    def attemptToImport(submission: CodeType) -> ModuleType:
        module = ModuleType("submission")

        exec(submission, vars(module))

        return module

    @staticmethod
    def applyInjectedCode(module: ModuleType, codeToInject: List[CodeType]) -> None:
        for code in codeToInject:
            exec(code, vars(module))

    @staticmethod
    def applyMocks(module: ModuleType, mocks: Dict[str, Optional[SingleFunctionMock]]) -> None:
        """
        This function applies the mocks to the student's submission at the module level.

        :param module: the module to apply mocks to
        :param mocks: the dictionary of mocks to apply

        :raises AttributeError: If a mock name cannot be resolved
        """
        for mockName, mock in mocks.items():
            if mock is None:
                continue

            if mock.spy:
                mock.setSpyFunction(getattr(module, mockName))

            setattr(module, mockName, mock)

    @staticmethod
    def getMethod(module: ModuleType, methodToRun: str) -> Callable:
        method: Optional[Callable] = getattr(module, methodToRun, None)

        if method is None:
            raise MissingFunctionDefinition(methodToRun)

        return method

    @staticmethod
    def runMethod(module: ModuleType, methodToRun: Callable[..., object],
                  parameters: List[Parameter]) -> RunMethodResult:
        processedParameters = tuple([parameter.get(module) for parameter in parameters])

        returnVal = methodToRun(*processedParameters)

        return {"return_val": returnVal, "parameters": processedParameters}

    @staticmethod
    def runMain(submission: CodeType) -> None:
        # Currently parameters are unsupported :(

        exec(submission, {'__name__': "__main__"})

    @staticmethod
    def resolveMocks(mocks: Dict[str, Optional[SingleFunctionMock]]) -> Dict[str, SingleFunctionMock]:
        subscribedMocks = [mockName for mockName, mock in mocks.items() if mock is None]
        # TODO logging

        for mock in subscribedMocks:
            splitName = mock.split('.')
            functionName = splitName[-1]

            try:
                mod = import_module(".".join(splitName[:-1]))
            except Exception as ex:
                raise ImportError(
                    f"Failed to import '{splitName}' during mock resolution. This is likely an autograder error.\n{str(ex)}")

            mockedFunction: Optional[SingleFunctionMock] = getattr(mod, functionName, None)

            if mockedFunction is None:
                raise ImportError(
                    f"Failed to locate '{functionName}' in '{splitName[:-1]}' during mock resolution. This is likely an autograder error.")

            mocks[mock] = mockedFunction

        return mocks

    @staticmethod
    def aggregateResults(runMethodResults: Optional[RunMethodResult],
                         mocks: Dict[str, SingleFunctionMock]) -> PythonTaskResult:
        return {
            "return_val": runMethodResults["return_val"] if runMethodResults is not None else None,
            "parameters": runMethodResults["parameters"] if runMethodResults is not None else None,
            "mocks": mocks,
        }


class PythonRunnerBuilder:
    INJECTED_PREFIX: Final[str] = "INJECTED_"

    def __init__(self: Builder, submission: PythonSubmission):
        self.submission: Final[CodeType] = submission.getExecutableSubmission()
        self.parameters: List[Parameter] = []
        self.mocks: Dict[str, Optional[SingleFunctionMock]] = {}
        self.injectedMethods: Dict[str, CodeType] = {}
        self.setupMethods: List[str] = []
        self.useModuleEntrypoint: bool = False
        self.functionEntrypoint: Optional[str] = None

    def addParameter(self: Builder, value: object = None, parameter: Optional[Parameter] = None) -> Builder:
        if parameter is None:
            parameter = Parameter(value)

        self.parameters.append(parameter)

        return self

    def addMock(self: Builder, name: str, mock: SingleFunctionMock) -> Builder:
        if name in self.mocks:
            raise InvalidRunner(f"Mock '{name}' has already been added to the runner.")

        if mock is None:
            raise InvalidRunner(f"Injected mock '{name}' must be defined! \nIf you are trying to subscribe to the output of a module mock, use 'subscribeToMock'")

        self.mocks[name] = mock

        return self

    def subscribeToMock(self: Builder, name: str) -> Builder:
        if name in self.mocks:
            raise InvalidRunner(f"Mock '{name}' has already been added to the runner.")

        self.mocks[name] = None

        return self

    def addInjectedCode(self: Builder, name: str, src: Optional[str] = None,
                        code: Optional[CodeType] = None) -> Builder:
        if name[0: len(PythonRunnerBuilder.INJECTED_PREFIX)] != PythonRunnerBuilder.INJECTED_PREFIX:
            raise InvalidRunner(
                f"Injected methods must be prefixed with '{PythonRunnerBuilder.INJECTED_PREFIX}'. Was '{name}'.")

        if src is not None and code is not None:
            raise InvalidRunner(f"Multiple injected method implementations provided. Expected either source *or* code.")

        if src is None and code is None:
            raise InvalidRunner(f"No injected method implementation provided. Expected *exactly* one.")

        if src is not None:
            try:
                code = compile(src, name, "exec")
            except SyntaxError as syntaxError:
                raise InvalidRunner(
                    f"Syntax Error when compiling injected method!\nEnsure that you are using dill.getsource(..., lstrip=True,...)\n{syntaxError}")
            except ValueError:
                raise InvalidRunner(f"Null Bytes detected when compiling injected method!")

        self.injectedMethods[name] = code

        return self

    def addSetupMethod(self: Builder, setupMethodName: str) -> Builder:
        self.setupMethods.append(setupMethodName)

        return self

    def setEntrypoint(self: Builder, module: bool = False, function: Optional[str] = None) -> Builder:
        if module and function is not None:
            raise InvalidRunner(f"Duplicate entrypoints defined. Only one can be defined!")

        self.useModuleEntrypoint = module
        self.functionEntrypoint = function

        return self

    def build(self) -> TaskRunner:
        if not self.functionEntrypoint and not self.useModuleEntrypoint:
            raise InvalidRunner(f"No entrypoint defined!")

        if self.functionEntrypoint and (
                PythonRunnerBuilder.INJECTED_PREFIX in self.functionEntrypoint and self.functionEntrypoint not in self.injectedMethods):
            raise InvalidRunner(f"Injected method '{self.functionEntrypoint}' has not been injected!")

        if self.useModuleEntrypoint and len(self.parameters) != 0:
            raise InvalidRunner(
                f"Incompatible options! No parameters can be defined when using module entrypoint. Use a environment mock of the 'sys' module instead.")

        taskRunner = TaskRunner(PythonSubmission)

        if self.useModuleEntrypoint:
            taskRunner.add(Task("main", PythonTaskLibrary.runMain, [lambda: self.submission]))
            taskRunner.add(Task("resolve_mocks", PythonTaskLibrary.resolveMocks, [lambda: self.mocks]))
            taskRunner.add(Task("results", PythonTaskLibrary.aggregateResults,
                                [lambda: None, lambda: taskRunner.getResult("resolve_mocks")]), isOverallResultTask=True)
            return taskRunner

        taskRunner.add(Task("import", PythonTaskLibrary.attemptToImport, [lambda: self.submission]))
        taskRunner.add(Task("injection", PythonTaskLibrary.applyInjectedCode,
                            [lambda: taskRunner.getResult("import"), lambda: self.injectedMethods.values()]))
        taskRunner.add(Task("apply_mock", PythonTaskLibrary.applyMocks, [lambda: taskRunner.getResult("import"), lambda: self.mocks]))
        methodOrder = self.setupMethods

        for method in methodOrder:
            taskRunner.add(Task(f"get_{method}", PythonTaskLibrary.getMethod,
                                [lambda: taskRunner.getResult("import"), lambda: method]))
            taskRunner.add(Task(f"run_{method}", PythonTaskLibrary.runMethod,
                                [lambda: taskRunner.getResult("import"), lambda: taskRunner.getResult(f"get_{method}"),
                                 lambda: []]))

        taskRunner.add(Task(f"get_{self.functionEntrypoint}", PythonTaskLibrary.getMethod,
                            [lambda: taskRunner.getResult("import"), lambda: self.functionEntrypoint]))
        taskRunner.add(Task(f"run_{self.functionEntrypoint}", PythonTaskLibrary.runMethod,
                            [lambda: taskRunner.getResult("import"),
                             lambda: taskRunner.getResult(f"get_{self.functionEntrypoint}"), lambda: self.parameters]))

        taskRunner.add(Task("resolve_mocks", PythonTaskLibrary.resolveMocks, [lambda: self.mocks]))
        taskRunner.add(Task("results", PythonTaskLibrary.aggregateResults,
                            [lambda: taskRunner.getResult(f"run_{self.functionEntrypoint}"),
                             lambda: taskRunner.getResult("resolve_mocks")]), isOverallResultTask=True)

        return taskRunner



from enum import Enum
from typing import Iterable, TypedDict, Tuple, Dict, Optional

from autograder_platform.TestingFramework.SingleFunctionMock import SingleFunctionMock


class FileTypeMap(Enum):
    TEST_FILES = 1
    PYTHON_FILES = 2
    REQUIREMENTS = 3

class PythonTaskResult(TypedDict):
    return_val: object
    parameters: Optional[Tuple[object, ...]]
    mocks: Dict[str, SingleFunctionMock]

class NoPyFilesError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Expected at least one `.py` file. Received 0.\n"
            "Are you writing your code in a file that ends with `.py`?"
        )

class MissingMainFileError(Exception):
    def __init__(self, expectedMains: Iterable[str], files: Iterable[str]) -> None:
        super().__init__(
            f"Expected file named {' or '.join(file for file in expectedMains)}. Received: {','.join(file for file in files)}"
        )

class TooManyFilesError(Exception):
    def __init__(self, files: Iterable[str]) -> None:
        super().__init__(
            f"Expected one `.py` file. Received: {', '.join(file for file in files)}\n"
            "Please delete extra `.py` files"
        )

class InvalidPackageError(Exception):
    def __init__(self, packageName: str, version: str):
        super().__init__(
            f"Unable to locate package, '{packageName}' at version "
            f"{version if version else 'any version'}"
        )

class InvalidRequirementsFileError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)



import os
from typing import Dict, Iterable, List, Optional


class MissingOutputDataException(Exception):
    def __init__(self, outputFileName):
        super().__init__("Output results are NULL.\n"
                         f"Failed to parse results in output buffer: {outputFileName}.\n"
                         f"Submission possibly crashed or terminated before harness could write to output buffer: {outputFileName}.\n"
                         f"Submission may have overrun buffer memory\n"
                         f"Likely causes: The presence of exit or quit in student's code; extra debugging print statements")

def filterStdOut(stdOut: Optional[List[str]]) -> Optional[List[str]]:
    """
    This function takes in a list representing the output from the program. It includes ALL output,
    so lines may appear as 'NUMBER> OUTPUT 3' where we only care about what is right after the OUTPUT statement
    This is adapted from John Henke's implementation

    :param _stdOut: The raw stdout from the program
    :returns: the same output with the garbage removed
    """

    if stdOut is None:
        return None

    filteredOutput: List[str] = []
    for line in stdOut:
        if "output " in line.lower():
            filteredOutput.append(line[line.lower().find("output ") + 7:])

    return filteredOutput

def detectFileSystemChanges(inFiles: Iterable[str], directoryToCheck: str) -> Dict[str, str]:
    files = [os.path.join(directoryToCheck, file) for file in os.listdir(directoryToCheck)]

    outputFiles: Dict[str, str] = {}

    # This ignores subfolders
    for path in files:
        if os.path.isdir(path):
            continue

        if "__" in os.path.basename(path):
            continue

        # ignore hidden files
        if os.path.basename(path)[0] == ".":
            continue

        if path in inFiles:
            continue

        outputFiles[os.path.basename(path)] = path

    return outputFiles



import importlib.util
from typing import Callable, Dict, List
import requests
import os
from autograder_platform.StudentSubmission.AbstractValidator import AbstractValidator
from autograder_platform.StudentSubmission.common import ValidationHook
from autograder_platform.StudentSubmissionImpl.Python.common import FileTypeMap, InvalidPackageError, InvalidRequirementsFileError, MissingMainFileError, NoPyFilesError, TooManyFilesError

class PythonFileValidator(AbstractValidator):

    @staticmethod
    def getValidationHook() -> ValidationHook:
        return ValidationHook.POST_LOAD

    def __init__(self, allowedMainNames: List[str]):
        super().__init__()
        self.allowedMainNames = allowedMainNames
        self.pythonFiles: Dict[FileTypeMap, List[str]] = {}
        self.looseMainMatchingAllowed: bool = False

    def setup(self, studentSubmission):
        submissionFiles = studentSubmission.getDiscoveredFileMap()
        self.looseMainMatchingAllowed = studentSubmission.getLooseMainMatchingEnabled()

        if studentSubmission.getTestFilesEnabled() and FileTypeMap.TEST_FILES in submissionFiles.keys():
            self.pythonFiles[FileTypeMap.TEST_FILES] = submissionFiles[FileTypeMap.TEST_FILES]

        if FileTypeMap.PYTHON_FILES not in submissionFiles:
            submissionFiles[FileTypeMap.PYTHON_FILES] = []

        self.pythonFiles[FileTypeMap.PYTHON_FILES] = [os.path.basename(file) for file in submissionFiles[FileTypeMap.PYTHON_FILES]]

    def run(self):
        if not self.pythonFiles[FileTypeMap.PYTHON_FILES]:
            self.addError(NoPyFilesError())
            return

        if self.looseMainMatchingAllowed and len(self.pythonFiles[FileTypeMap.PYTHON_FILES]) > 1:
            self.addError(TooManyFilesError(self.pythonFiles[FileTypeMap.PYTHON_FILES]))
            return

        if self.looseMainMatchingAllowed:
            return

        mainNameFilter: Callable[[str], bool] = lambda x: x in self.allowedMainNames

        filteredFiles = list(filter(mainNameFilter, self.pythonFiles[FileTypeMap.PYTHON_FILES]))

        if not filteredFiles:
            self.addError(MissingMainFileError(self.allowedMainNames, self.pythonFiles[FileTypeMap.PYTHON_FILES]))
            return
        
        if len(filteredFiles) != 1:
            self.addError(TooManyFilesError(filteredFiles))

class RequirementsValidator(AbstractValidator):
    @staticmethod
    def getValidationHook() -> ValidationHook:
        return ValidationHook.POST_LOAD

    def __init__(self):
        super().__init__()
        self.requirements: List[str] = []
        self.submissionBase: str = ""

    def setup(self, studentSubmission):
        self.submissionBase = studentSubmission.getSubmissionRoot()
        files = studentSubmission.getDiscoveredFileMap()
        if FileTypeMap.REQUIREMENTS not in files:
            files[FileTypeMap.REQUIREMENTS] = []

        self.requirements = files[FileTypeMap.REQUIREMENTS]

    def run(self):
        if not self.requirements:
            return

        if len(self.requirements) != 1:
            self.addError(InvalidRequirementsFileError(
                    "Too many requirements.txt files.\n"
                    f"Expected 1, received {len(self.requirements)}"
                )
            )
        for file in self.requirements:
            if file == os.path.join(self.submissionBase, "requirements.txt"):
                continue

            self.addError(InvalidRequirementsFileError(
                    "Invalid location for requirements file.\n"
                    f"Should be {os.path.join(self.submissionBase, 'requirements.txt')}\n"
                    f"But was {file}"
                )
            )

class PackageValidator(AbstractValidator):

    PYPI_BASE = "https://pypi.org/pypi/"

    @staticmethod
    def getValidationHook() -> ValidationHook:
        return ValidationHook.PRE_BUILD

    def __init__(self):
        super().__init__()
        self.packages: Dict[str, str] = {}

    def setup(self, studentSubmission):
        self.packages = studentSubmission.getExtraPackages()

    def run(self):

        for package, version in self.packages.items():
            if importlib.util.find_spec(package) != None:
                continue

            url = self.PYPI_BASE + package + "/"

            if version:
                url += version + "/"

            url += "json"

            if requests.get(url=url).status_code == 200:
                continue

            self.addError(InvalidPackageError(package, version))


        




"""
This module provides an interface for running student submissions as processes.

This module utilizes shared memory to share data from the parent and the child. The way it is currently implemented
guarantees that no conflicts will occur with the pattern. This is because the parent waits for the child to complete
before attempting to even connect to the object.

:author: Gregory Bell
:date: 3/7/23
"""

from typing import Any, Dict, Optional, TextIO, Tuple, List, Union
from autograder_platform.Executors.Environment import ExecutionEnvironment, Results

from autograder_platform.StudentSubmission.ISubmissionProcess import ISubmissionProcess

import dill
import multiprocessing
import multiprocessing.shared_memory as shared_memory
import os
import sys
from io import StringIO

from autograder_platform.Executors.common import MissingOutputDataException, detectFileSystemChanges, filterStdOut
from autograder_platform.StudentSubmissionImpl.Python.common import PythonTaskResult
from autograder_platform.Tasks.TaskRunner import TaskRunner
from autograder_platform.TestingFramework.SingleFunctionMock import SingleFunctionMock
from autograder_platform.StudentSubmissionImpl.Python.PythonEnvironment import PythonEnvironment, PythonResults
from autograder_platform.StudentSubmissionImpl.Python.AbstractPythonImportFactory import AbstractModuleFinder

dill.Pickler.dumps, dill.Pickler.loads = dill.dumps, dill.loads  # type: ignore
multiprocessing.reduction.dump = dill.dump  # type: ignore

class StudentSubmissionProcess(multiprocessing.Process):
    """
    This class extends multiprocessing.Process to provide a simple way to run student submissions.
    This class runs the 'runner' that it receives from the parent in a separate process and shares all data collected
    (returns, stdout, and exceptions) with the parent who is able to unpack it. Currently, stderror is ignored.

    Depending on the platform the process will be spawned using either the 'fork' or 'spawn' methods.
    Windows only supports the 'spawn' method and the 'spawn' method is preferred on macOS. This is why we aren't using
    pipes; when a process is created via the 'spawn' method, a new interpreter is started which doesn't share resources
    in the same way that a child process created via the 'fork' method. Thus, pipes fail on non Linux platforms.

    The other option of using multiprocessing.Pipe was also considered, but it ran in to similar issues as the
    traditional pipe method. There is a way to make it work, but it would involve overriding the print function to
    utilize the Pipe.send function, but that was a whole host of edge cases that I did not want to consider and also
    seemed like an easy way to confuse students as to why their print function behaved weird in the autograder, but not
    on their computers.

    Please be aware that the data does not generally persist after the child is started as it is a whole new context on
    some platforms. So data must be shared via some other mechanism. I choose multiprocessing.shared_memory due to the
    limitations listed above.

    This class is designed to a one-size-fits-all type of approach for actually *running* student submissions as I
    wanted to avoid the hodgepodge of unmaintainable that was the original autograder while still affording the
    flexibility required by the classes that will utilize it.
    """

    def __init__(self, runner: TaskRunner, executionDirectory: str, importHandlers: List[AbstractModuleFinder],
                 timeout: int = 10):
        """
        This constructs a new student submission process with the name "Student Submission".

        It sets the default names for shared input and output.
        Those should probably be updated.

        :param runner: The submission runner to be run in a new process. Can be any callable object (lamda, function,
        etc). If there is a return value it will be shared with the parent.


        :param executionDirectory: The directory that this process should be running in. This is to make sure that all
        data is isolated for each run of the autograder.

        :param timeout: The _timeout for join. Basically, it will wait *at most* this amount of time for the child to
        terminate. After this period passes, the child must be killed by the parent.
        """
        super().__init__(name="Student Submission")
        self.runner: TaskRunner = runner
        self.inputDataMemName: str = ""
        self.outputDataMemName: str = ""
        self.executionDirectory: str = executionDirectory
        self.importHandlers: List[AbstractModuleFinder] = importHandlers
        self.timeout: int = timeout

    def setInputDataMemName(self, inputSharedMemName):
        """
        Updates the input data memory name from the default

        :param inputSharedMemName: The shared memory name (see :ref:`multiprocessing.shared_memory`) for stdin.
        The data at this location is stored as a list
        and must be processed into a format understood by ``StringIO``.
        The data must exist before the child is started.
        """
        self.inputDataMemName: str = inputSharedMemName

    def setOutputDataMenName(self, outputDataMemName):
        """
        Updates the output data memory name from the default.

        :param outputDataMemName: The shared memory name (see :ref:`multiprocessing.shared_memory`) for exceptions and
        return values.
        This is created by the child and will be connected to by the parent once the child exits.
        """
        self.outputDataMemName = outputDataMemName

    def _setup(self) -> None:
        """
        Sets up the child input output redirection. The stdin is read from the shared memory object defined in the parent
        with the name ``self.inputDataMemName``. The stdin is formatted with newlines so that ``StringIO`` is able to
        work with it.

        This method also moves the process to the execution directory

        stdout is also redirected here, but because we don't care about its contents, we just overwrite it completely.

        This method also injects whatever import MetaPathFinders
        """
        # This may error? so we are going to catch it and log the error
        try:
            os.chdir(self.executionDirectory)
        except OSError as ex:  # pragma: no coverage
            print(f"ERROR: Failed to change directory to sandbox folder.\n{ex}", file=sys.stderr)  # pragma: no coverage


        sys.path.append(os.getcwd())

        for importHandler in self.importHandlers:
            sys.meta_path.insert(0, importHandler)

            mods = importHandler.getModulesToReload()
            for mod in mods:
                if mod not in sys.modules:
                    continue

                del sys.modules[mod]

        sharedInput = shared_memory.SharedMemory(self.inputDataMemName)
        deserializedData = dill.loads(sharedInput.buf.tobytes())
        # Reformat the stdin so that we
        sys.stdin = StringIO("".join([line + "\n" for line in deserializedData]))

        sys.stdout = StringIO()

    def _teardown(self, stdout: Union[StringIO, TextIO], exception: Optional[Exception],
                  returnValue: object, parameters: Optional[Tuple[object, ...]],
                  mocks: Optional[Dict[str, Optional[SingleFunctionMock]]]) -> None:
        """
        This function takes the results from the child process and serializes them.
        Then is stored in the shared memory object that the parent is able to access.

        :param stdout: The raw io from the stdout.
        :param exception: Any exceptions that were thrown
        :param returnValue: The return value from the function
        :param mocks: The mocks from the submission after they have been hydrated
        """

        if isinstance(stdout, TextIO):
            stdout.seek(0)
            stdout = StringIO(stdout.read())

        # Pickle both the exceptions and the return value
        dataToSerialize: Dict[str, Any] = {
            "stdout": stdout.getvalue().splitlines(),
            "parameters": parameters,
            "return_val": returnValue,
            "exception": exception,
            "impl_results": {
                "mocks": mocks,
            },
        }

        for importHandler in self.importHandlers:
            sys.meta_path.remove(importHandler)

        serializedData = dill.dumps(dataToSerialize, dill.HIGHEST_PROTOCOL)
        sharedOutput = shared_memory.SharedMemory(self.outputDataMemName)

        if sharedOutput.size < sys.getsizeof(serializedData):
            print(f"FATAL ERROR: Submission generated output is LARGER than buffer size of {sharedOutput.size} bytes. Output size is {sys.getsizeof(serializedData)} bytes!", file=sys.stderr)
            sharedOutput.close()
            return

        sharedOutput.buf[:len(serializedData)] = serializedData
        sharedOutput.close()

    def run(self):
        self._setup()

        exception: Optional[Exception] = None

        results: PythonTaskResult = self.runner.run()  # type: ignore

        if not self.runner.wasSuccessful():
            exceptions = self.runner.getAllErrors()
            if exceptions:
                exception = exceptions[0]

        if results is None:
            results = {
                "return_val": None,
                "parameters": None,
                "mocks": {},
            }

        self._teardown(sys.stdout, exception, results["return_val"], results["parameters"], results["mocks"])

    def join(self, *args, **kwargs):
        multiprocessing.Process.join(self, timeout=self.timeout)

    def terminate(self):
        # SigKill - cant be caught
        multiprocessing.Process.kill(self)
        # Checks to see if we are killed and cleans up process
        multiprocessing.Process.terminate(self)
        # Clean up the zombie
        multiprocessing.Process.join(self, timeout=0)


class RunnableStudentSubmission(ISubmissionProcess):

    def __init__(self):
        self.inputSharedMem: Optional[shared_memory.SharedMemory] = None
        self.outputSharedMem: Optional[shared_memory.SharedMemory] = None

        self.runner: Optional[TaskRunner] = None
        self.executionDirectory: str = "."
        self.studentSubmissionProcess: Optional[StudentSubmissionProcess] = None
        self.exception: Optional[Exception] = None
        self.outputData: Dict[str, Any] = {}
        self.timeoutOccurred: bool = False
        self.timeoutTime: int = 0
        self.bufferSize: int = 0

    def setup(self, environment: ExecutionEnvironment[PythonEnvironment, PythonResults], runner: TaskRunner):
        """
        Description
        ---

        This function allocates the shared memory that will be passed to the student's submission

        Setting up the data here then tearing it down in the ref:`RunnableStudentSubmission.cleanup` fixes
        the issue with windows GC cleaning up the memory before we are done with it as there will be at least one
        active hook for each memory resource til ``cleanup`` is called.


        """
        self.studentSubmissionProcess = \
            StudentSubmissionProcess(runner, environment.sandbox_location,
                                     environment.impl_environment.import_loader,
                                     environment.timeout)

        self.bufferSize = environment.impl_environment.buffer_size

        if self.bufferSize <= 0:
            raise AttributeError("INVALID STATE: Buffer size is ZERO. No data can be collected from the student's submission.")

        self.inputSharedMem = shared_memory.SharedMemory(create=True, size=self.bufferSize)
        self.outputSharedMem = shared_memory.SharedMemory(create=True, size=self.bufferSize)

        self.studentSubmissionProcess.setInputDataMemName(self.inputSharedMem.name)
        self.studentSubmissionProcess.setOutputDataMenName(self.outputSharedMem.name)

        serializedStdin = dill.dumps(environment.stdin, dill.HIGHEST_PROTOCOL)

        self.inputSharedMem.buf[:len(serializedStdin)] = serializedStdin

        self.timeoutTime = environment.timeout

    def run(self):
        if self.studentSubmissionProcess is None:
            raise AttributeError("Process has not be initialized!")

        self.studentSubmissionProcess.start()

        self.studentSubmissionProcess.join()

        if self.studentSubmissionProcess.is_alive():
            self.studentSubmissionProcess.terminate()
            self.timeoutOccurred = True

    def _deallocate(self):
        if self.inputSharedMem is None or self.outputSharedMem is None:
            return

        # `close` closes the current hook
        self.inputSharedMem.close()
        # `unlink` tells the gc that it is ok to clean up this resource
        #  On windows, `unlink` is a noop
        self.inputSharedMem.unlink()

        self.outputSharedMem.close()
        self.outputSharedMem.unlink()

    def cleanup(self):
        """
        This function cleans up the shared memory object by closing the parent hook and then unlinking it.

        After it is unlinked, the python garbage collector cleans it up.
        On windows, the GC runs as soon as the last hook is closed and `unlink` is a noop
        """

        if self.inputSharedMem is None or self.outputSharedMem is None:
            return

        if self.timeoutOccurred:
            self.exception = TimeoutError(f"Submission timed out after {self.timeoutTime} seconds")
            self._deallocate()
            return

        # This prolly isn't the best memory wise, but according to some chuckle head on reddit, this is superfast
        outputBytes = self.outputSharedMem.buf.tobytes()

        if outputBytes == bytearray(self.bufferSize):
            self.exception = MissingOutputDataException(self.outputSharedMem.name)
            self._deallocate()
            return

        deserializedData: Dict[str, Any] = dill.loads(outputBytes)

        self.outputData = deserializedData

        self._deallocate()

    def populateResults(self, environment: ExecutionEnvironment):
        if not self.outputData:
            self.outputData = {
                "stdout": None,
                "parameters": None,
                "return_val": None,
                "exception": self.exception,
                "impl_results": {
                    "mocks": None,
                },
            }

        self.outputData["file_out"] = detectFileSystemChanges(environment.files.values(), environment.sandbox_location)
        self.outputData["stdout"] = filterStdOut(self.outputData["stdout"])

        if "impl_results" in self.outputData:
            self.outputData["impl_results"] = PythonResults(**self.outputData["impl_results"])

        environment.resultData = Results(**self.outputData)

    @classmethod
    def processAndRaiseExceptions(cls, environment: ExecutionEnvironment):
        if environment.resultData is None:
            return

        exception = environment.resultData.exception

        if exception is None:
            return

        errorMessage = f"Submission execution failed due to an {type(exception).__qualname__} exception.\n" + str(
            exception)

        if isinstance(exception, EOFError):
            errorMessage += "\n" \
                            "Do you have the correct number of input statements?\n" \
                            "Are your loops terminating correctly?\n" \
                            "Is all your code in the if __name__ == __main__ block if you are using functions?"

        raise AssertionError(errorMessage)



import os

from importlib.machinery import ModuleSpec
from types import ModuleType, CodeType
from typing import Dict, Optional, List
from importlib.abc import Loader
from importlib.util import spec_from_file_location

from autograder_platform.StudentSubmissionImpl.Python.AbstractPythonImportFactory import AbstractModuleFinder

class ModuleFinder(AbstractModuleFinder):
    def __init__(self) -> None:
        self.knownModules: Dict[str, str] = {}
        self.modulesToReload: List[str] = []

    def addModule(self, fullname, path):
        for mod in fullname.split('.'):
            self.knownModules[mod] = path
            self.modulesToReload.append(mod)

    def find_spec(self, fullname, path, target=None):
        if fullname not in self.knownModules:
            return None
        
        return spec_from_file_location(fullname, self.knownModules[fullname], loader=ModuleLoader(self.knownModules[fullname]))
    
    def getModulesToReload(self) -> List[str]:
        return self.modulesToReload


class ModuleLoader(Loader):
    def __init__(self, filename):
        self.filename = filename

    def create_module(self, spec: ModuleSpec) -> Optional[ModuleType]:
        return None
    
    def exec_module(self, module: ModuleType) -> None:
        if not os.path.exists(self.filename):
            raise ImportError(f"Should be able to open {self.filename}, but was unable to locate file!")
        with open(self.filename) as r:
            data = r.read()
        
        compiledImport: CodeType = compile(data, self.filename, "exec")
        exec(compiledImport, vars(module))
    
class PythonFileImportFactory:
    moduleFinder: ModuleFinder = ModuleFinder()

    @classmethod
    def registerFile(cls, pathToFile: str, importName: str):
        if cls.moduleFinder == None:
            raise AttributeError("Invalid State: Module finder is none")
        if "addModule" in vars(cls.moduleFinder):
            raise AttributeError("Invalid ModuleFinder for registration")

        cls.moduleFinder.addModule(importName, pathToFile)
    
    @classmethod
    def buildImport(cls):
        return cls.moduleFinder




from typing import Dict, Generic, List, TypeVar
from abc import ABC, abstractmethod, abstractproperty

T = TypeVar("T")

class BaseSchema(Generic[T], ABC):

    @abstractmethod
    def validate(self, data: Dict) -> Dict:
        raise NotImplementedError()

    @abstractmethod
    def build(self, data: Dict) -> T:
        raise NotImplementedError()

class InvalidConfigException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class MissingParsingLibrary(Exception):
    def __init__(self, library, parserName) -> None:
        super().__init__(f"Missing {library}. Required for {parserName}")



import getpass
import os
import sys
from zipfile import ZipFile


def buildZipName() -> str:
    """
    This function generates the zip file name and sets a fallback if the user's name can't be determined
    :return: the generated zip file name
    """
    DEFAULT_NAME: str = "csci_128_student"
    zipName: str = ""

    try:
        # Get user can throw an exception if it fails to look up the username in the password database
        userName = getpass.getuser().lower()
        # replace spaces with underscore bc spaces make me sad
        userName = userName.replace(" ", "_")

        zipName += userName

    except KeyError:
        print(f"Can't automatically determine user name. Defaulting to {DEFAULT_NAME}")
        zipName += DEFAULT_NAME

    zipName += "-submission"
    zipName += ".zip"

    return zipName


def addFolderToZip(_currentZipBuffer: ZipFile, _directoryToAdd: str) -> None:
    """
    This function allows us to recursively descend through the directory structure to
    collect all of a student's submitted files
    :param _currentZipBuffer: the zip file that is currently being written.
    :param _directoryToAdd: the directory that needs to be added
    """

    # Ignore hidden directories
    if _directoryToAdd[0] == ".":
        return

    # Ignore directories python directories
    if "__" in _directoryToAdd:
        return

    print(f"\tEntering {_directoryToAdd}...")

    for file in os.listdir(_directoryToAdd):
        if os.path.isfile(_directoryToAdd + file) and file[-3:] == ".py":
            print(f"\tAdding {_directoryToAdd + file}...")
            _currentZipBuffer.write(_directoryToAdd + file)
        elif os.path.isfile(_directoryToAdd + file):
            print(f"\tIgnoring {_directoryToAdd + file}...")
        else:
            addFolderToZip(_currentZipBuffer, _directoryToAdd + file + "/")


def generateZipFile(_submissionDirectory: str) -> None:
    """
    This function generates the zip file needed for gradescope.
    :param _submissionDirectory: the directory that the students work is in
    """
    GREEN_COLOR: str = u"\u001b[32m"
    RESET_COLOR: str = u"\u001b[0m"

    print("Generating gradescope upload...")

    zipName: str = buildZipName()

    with ZipFile(zipName, 'w') as submissionZip:
        os.chdir(_submissionDirectory)
        for file in os.listdir("."):
            if os.path.isfile(file) and file[-3:] == ".py":
                print(f"\tAdding {_submissionDirectory + file} to zip...")
                submissionZip.write(file)
            elif os.path.isfile(file):
                print(f"\tIgnoring {_submissionDirectory + file}...")
            else:
                addFolderToZip(submissionZip, file + "/")

    print("...Done.")
    print(f"\n\n{GREEN_COLOR}Submit {zipName} to Gradescope under the corresponding assignment.{RESET_COLOR}")



def tool():  # pragma: no cover
    submissionDirectory = "student_work"

    if len(sys.argv) == 2:
        submissionDirectory = sys.argv[1]

    # need to make sure to that we have a / at the end of the path
    if submissionDirectory[-1:] != '/':
        submissionDirectory += "/"

    generateZipFile(submissionDirectory)


if __name__ == "__main__":
    tool()



from enum import Enum
from typing import List, Any


class ValidationHook(Enum):
    PRE_LOAD = 1
    POST_LOAD = 2
    PRE_BUILD = 3
    POST_BUILD = 4
    VALIDATION = 5


class MissingFunctionDefinition(Exception):
    def __init__(self, functionName: str):
        super().__init__(
            f"Failed to find function with name: {functionName}.\n"
            "Are you missing the function definition?"
        )

        self.functionName = functionName

    # https://stackoverflow.com/questions/16244923/how-to-make-a-custom-exception-class-with-multiple-init-args-pickleable
    # Basically - reduce has to return something that we constuct the og class from
    def __reduce__(self):
        # Need to be (something,) so that it actually gets processed as a tuple in the pickler
        return (MissingFunctionDefinition, (self.functionName,))


class InvalidTestCaseSetupCode(Exception):
    def __init__(self, *args):
        super().__init__(
            "Failed to find 'autograder_setup' function to run.\n"
            "Ensure that your setup code contains a 'autograder_setup' function.\n"
            "This is an autograder error."
        )


class StudentSubmissionException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ValidationError(Exception):
    @staticmethod
    def combineErrorMessages(exceptions: List[Exception]) -> str:
        msg = ""
        for i, ex in enumerate(exceptions):
            msg += f"{i + 1}. {type(ex).__qualname__}: {ex}\n"

        return msg

    def __init__(self, exceptions: List[Exception]):
        msg = self.combineErrorMessages(exceptions)

        super().__init__("Validation Errors:\n" + msg)


class InvalidRunner(Exception):
    def __init__(self, msg):
        super().__init__("Invalid Runner State: " + msg)



import importlib
import os
from typing import Dict, Generic, List, Optional as OptionalType, TypeVar, Any
from dataclasses import dataclass

from schema import And, Optional, Or, Regex, Schema, SchemaError

from autograder_platform.config.common import BaseSchema, MissingParsingLibrary, InvalidConfigException


@dataclass(frozen=True)
class BuildConfiguration:
    """
    Build Configuration
    ===================
    
    This class defines the build options when building new or existing autograders
    """
    use_starter_code: bool
    """Whether the starter code should be pulled into the student autograder"""
    use_data_files: bool
    """Whether datafiles should be pulled into the autograder"""
    allow_private: bool
    """Whether the private tests and datafiles should be kept private"""
    build_student: bool
    """If we should build the student autograder"""
    build_gradescope: bool
    """If we should build the gradescope autograder"""
    build_prairie_learn: bool
    """If we should build the PrairieLearn autograder"""
    data_files_source: str
    """The folder that contains the datafiles for use with the autograder"""
    starter_code_source: str
    """The path for the starter code that should be provided to the student"""
    student_work_folder: str
    """The folder that should be created for the student"""
    private_tests_regex: str
    """The pattern that should be used to identify private tests"""
    public_tests_regex: str
    """The pattern that should be used to identify public tests"""


@dataclass(frozen=True)
class PythonConfiguration:
    """
    Python Configuration
    ====================

    This class defines extra parameters for when the autograder is running in Python
    """
    extra_packages: List[Dict[str, str]]
    """
    The extra packages that should be added to the autograder on build.
    Must be stored in 'package_name': 'version'. Similar to requirements.txt 
    """
    buffer_size: int
    """
    The size of the output buffer when the autograder runs
    """


@dataclass(frozen=True)
class CConfiguration:
    """
    C/C++/C-Like Configuration
    ==========================

    This defines the extra parameters for when the autograder is running for c like languages
    """

    use_makefile: bool
    """
    If a makefile should be used for building
    """
    clean_target: str
    """
    The target that should be used to clean. Invoked as `make {clean_target}`
    """
    submission_name: str
    """
    The file name that should be executed
    """


@dataclass(frozen=True)
class BasicConfiguration:
    """
    Basic Configuration
    ===================

    This class defines the basic autograder configuration
    """
    impl_to_use: str
    """The StudentSubmission Implementation to use"""
    student_submission_directory: str
    """The folder that the student submission is in"""
    autograder_version: str
    """The autograder version tag. Autograder will be kept at this version"""
    test_directory: str
    """The directory that the tests are in"""
    enforce_submission_limit: bool
    """Whether or not the submission limit should be enforced"""
    submission_limit: int
    """The max number of submissions that a student has"""
    take_highest: bool
    """If we should take the highest of all the valid scores"""
    allow_extra_credit: bool
    """If scores greater than ``perfect_score`` should be respected"""
    perfect_score: int
    """The max score with out extra credit that a student can get"""
    max_score: int
    """
    The max score that students can get with extra credit. 
    Points greater than this will not be honored.
    """
    python: OptionalType[PythonConfiguration] = None
    """Extra python spefic configuration. See :ref:`PythonConfiguration` for options"""
    c: OptionalType[CConfiguration] = None
    """Extra C/C-like spefic configuration. See :ref:`CConfiguration` for options"""


@dataclass(frozen=True)
class AutograderConfiguration:
    """
    Autograder Configuration
    ========================

    The root object for the autograder configuration.
    Comprised of sub objects. 
    See :ref:`BasicConfiguration` and :ref:`BuildConfiguration` for more information.
    """
    assignment_name: str
    """The assignment name. IE: `ConstructionSite`"""
    semester: str
    """The semester that this assignment is being offered. IE: F99 -> Fall 1999"""
    autograder_root: str
    """The autograder's root directory where the config.toml file is located"""
    config: BasicConfiguration
    """The basic settings for the autograder. See :ref:`BasicConfiguration` for options."""
    build: BuildConfiguration
    """The build configuration for the autograder. See :ref:`BuildConfiguration` for options."""


class AutograderConfigurationSchema(BaseSchema[AutograderConfiguration]):
    """
    Autograder Configuration Schema
    ===============================

    This class defines the format agnostic schema required for the autograder.
    This class is able to validate and build a config. 
    Configs are expected to be provided as a dictionary; hence that agnostic nature of the schema.

    This class builds to: ref:`AutograderConfiguration` for easy typing.
    """
    IMPL_SOURCE = "StudentSubmissionImpl"

    @staticmethod
    def validateImplSource(implName: str) -> bool:
        try:
            importlib.import_module(f"autograder_platform.{AutograderConfigurationSchema.IMPL_SOURCE}.{implName}")
        except ImportError:
            return False
        return True

    def __init__(self):
        self.currentSchema: Schema = Schema(
            {
                "assignment_name": And(str, Regex(r"^(\w+-?)+$")),
                "semester": And(str, Regex(r"^(F|S|SUM)\d{2}$")),
                Optional("autograder_root", default="."): And(os.path.exists, os.path.isdir, lambda path: "config.toml" in os.listdir(path)),
                "config": {
                    "impl_to_use": And(str, AutograderConfigurationSchema.validateImplSource),
                    Optional("student_submission_directory", default="."): And(str, os.path.exists, os.path.isdir),
                    "autograder_version": And(str, Regex(r"\d+\.\d+\.\d+")),
                    "test_directory": And(str, os.path.exists),
                    "enforce_submission_limit": bool,
                    Optional("submission_limit", default=1000): And(int, lambda x: x >= 1),
                    Optional("take_highest", default=True): bool,
                    Optional("allow_extra_credit", default=False): bool,
                    "perfect_score": And(int, lambda x: x >= 1),
                    "max_score": And(int, lambda x: x >= 1),
                    Optional("python", default=None): Or({
                        Optional("extra_packages", default=lambda: []): [{
                            "name": str,
                            "version": str,
                        }],
                        Optional("buffer_size", default=2 ** 20): And(int, lambda x: x >= 2 ** 20)
                    }, None),
                    Optional("c", default=None): Or({
                        "use_makefile": bool,
                        "clean_target": str,
                        "submission_name": And(str, lambda x: len(x) >= 1)
                    }, None),
                },
                "build": {
                    "use_starter_code": bool,
                    "use_data_files": bool,
                    Optional("allow_private", default=True): bool,
                    Optional("data_files_source", default=None): str,
                    Optional("starter_code_source", default=None): str,
                    "build_student": bool,
                    "build_gradescope": bool,
                    Optional("build_prairie_learn", default=False): bool,
                    Optional("student_work_folder", default="student_work"): str,
                    Optional("private_tests_regex", default=r"^test_private_?\w*\.py$"): str,
                    Optional("public_tests_regex", default=r"^test_?\w*\.py$"): str,
                }
            },
            ignore_extra_keys=False, name="ConfigSchema"
        )

    def validate(self, data: Dict) -> Dict:
        """
        Description
        ---
        This method validates the provided data against the schema.

        If it is valid, then it will be returned. Otherwise, an ``InvalidConfigException`` will be raised.

        :param data: The data to validate
        :return: The data if it is able to be validated
        """
        validated = {}
        try:
            validated = self.currentSchema.validate(data)
        except SchemaError as schemaError:
            raise InvalidConfigException(str(schemaError))

        impl_to_use = validated["config"]["impl_to_use"].lower()

        if impl_to_use not in validated["config"] or validated["config"][impl_to_use] is None:
            raise InvalidConfigException(f"Missing Implementation Config for config.{impl_to_use}")

        if validated["build"]["use_starter_code"] and validated["build"]["starter_code_source"] is None:
            raise InvalidConfigException("Missing starter code file location")

        if validated["build"]["use_data_files"] and validated["build"]["data_files_source"] is None:
            raise InvalidConfigException("Missing directory for data files!")

        return validated

    def build(self, data: Dict) -> AutograderConfiguration:
        """
        Description
        ---
        This method builds the provided data into the known config format.

        In this case, it builds into the ``AutograderConfiguration`` format.
        Data should be validated before calling this method as it uses dictionary expandsion to populate the config objects.

        Doing this allows us to have a strongly typed config format to be used later in the autograder.
        """
        if data["config"]["python"] is not None:
            data["config"]["python"] = PythonConfiguration(**data["config"]["python"])

        if data["config"]["c"] is not None:
            data["config"]["c"] = CConfiguration(**data["config"]["c"])

        data["config"] = BasicConfiguration(**data["config"])
        data["build"] = BuildConfiguration(**data["build"])

        return AutograderConfiguration(**data)


# Using generics as PyRight and mypy are able to infer what `T` should be from the Schema
#  as it inherits from BaseSchema
T = TypeVar("T")
Builder = TypeVar("Builder", bound="AutograderConfigurationBuilder[Any]")


class AutograderConfigurationBuilder(Generic[T]):
    """
    AutograderConfigurationBuilder
    ==============================

    This class currently doesn't do much.
    It allows a schema (that inherits from :ref:`BaseSchema`) to passed in to use to validate and build.
    However, it assumes the ``AutograderConfigurationSchema`` will be used.

    This allows loading from currently only toml files, but is very easy to expand to different file formats if needed.

    ``.build`` should always be the last thing called.

    In the future, configuration will be allowed with this builder, but I would need to see the use case.
    """
    DEFAULT_CONFIG_FILE = "./config.toml"

    def __init__(self, configSchema: BaseSchema[T] = AutograderConfigurationSchema()):
        self.schema: BaseSchema[T] = configSchema
        self.data: Dict = {}

    def fromTOML(self: Builder, file=DEFAULT_CONFIG_FILE, merge=True) -> Builder:
        """
        Attempt to load the autograder config from the TOML config file.
        This file is assumed to be located in the same directory as the actual test cases
        """
        try:
            from tomli import load
        except ModuleNotFoundError:
            raise MissingParsingLibrary("tomlkit", "AutograderConfigurationBuilder.fromTOML")

        with open(file, 'rb') as rb:
            self.data = load(rb)

        return self

    # Really easy to add support for other file formats.
    # YAML or JSON would work as well

    @staticmethod
    def _createKeyIfDoesntExist(source: Dict[str, Any], key: str):
        if key in source:
            return

        source[key] = {}

    def setStudentSubmissionDirectory(self: Builder, studentSubmissionDirectory: OptionalType[str]) -> Builder:
        if studentSubmissionDirectory is None:
            return self

        self._createKeyIfDoesntExist(self.data, "config")

        self.data["config"]["student_submission_directory"] = studentSubmissionDirectory

        return self

    def setTestDirectory(self: Builder, testDirectory: str) -> Builder:
        self._createKeyIfDoesntExist(self.data, "config")

        self.data["config"]["test_directory"] = testDirectory

        return self

    def setAutograderRoot(self: Builder, autograderDirectory: str) -> Builder:
        self.data["autograder_root"] = autograderDirectory

        return self

    def build(self) -> T:
        self.data = self.schema.validate(self.data)
        return self.schema.build(self.data)


class AutograderConfigurationProvider:
    """
    AutograderConfigurationProvider
    ===============================

    This class allows access to the same config across the entire program.
    This is using a similar pattern to singletons, however, it's a bit better as it's a separate provider.
    """
    config: OptionalType[AutograderConfiguration] = None

    @classmethod
    def get(cls) -> AutograderConfiguration:
        if cls.config is None:
            raise AttributeError("Configuration has not been set!")

        return cls.config

    @classmethod
    def set(cls, config: AutograderConfiguration):
        if cls.config is not None:
            raise AttributeError("Configuration has already been set!")

        cls.config = config

    @classmethod
    def reset(cls):
        cls.config = None



from autograder_platform.StudentSubmission.SubmissionProcessFactory import SubmissionProcessFactory
from autograder_platform.StudentSubmissionImpl.Python.PythonEnvironment import PythonEnvironment, configMapper

from autograder_platform.StudentSubmissionImpl.Python.PythonSubmission import PythonSubmission
from autograder_platform.StudentSubmissionImpl.Python.PythonSubmissionProcess import RunnableStudentSubmission

SubmissionProcessFactory.register(PythonSubmission, RunnableStudentSubmission, PythonEnvironment, configMapper)



import os
import re
import sys
import subprocess
from types import CodeType
from typing import Dict, Iterable, List, Optional, TypeVar
from autograder_platform.StudentSubmission.AbstractStudentSubmission import AbstractStudentSubmission
from autograder_platform.StudentSubmissionImpl.Python.PythonValidators import PythonFileValidator, PackageValidator, RequirementsValidator
from autograder_platform.StudentSubmissionImpl.Python.common import FileTypeMap

Builder = TypeVar("Builder", bound="PythonSubmission")


def filterSearchResults(path: str) -> bool:
    # ignore hidden files
    if path[0] == '.':
        return False

    # ignore python cache files
    if "__pycache__" in path:
        return False

    if " " in path:
        return False

    return True


class PythonSubmission(AbstractStudentSubmission[CodeType]):
    ALLOWED_STRICT_MAIN_NAMES = ["main.py", "submission.py"]

    # we are not allowing any files with a space in them
    PYTHON_FILE_REGEX: re.Pattern = re.compile(r"^(\w|-)+\.py$")
    # the requirements.txt file MUST be in the root of the student's submission
    REQUIREMENTS_REGEX: re.Pattern = re.compile(r"^requirements\.txt$")
    # test files follow similar rules to python files, but must start with 'test'
    TEST_FILE_REGEX: re.Pattern = re.compile(r"^test(\w|-)*\.py$")
    # this allows versioned and non versioned packages, but disallows local packages
    REQUIREMENTS_LINE_REGEX: re.Pattern = re.compile(r"^(\w|-)+(==)?(\d+\.?){0,3}$")

    def __init__(self):
        super().__init__()

        self.testFilesEnabled: bool = False
        self.requirementsEnabled: bool = False
        self.looseMainMatchingEnabled: bool = False

        self.discoveredFileMap: Dict[FileTypeMap, List[str]] = {}

        self.extraPackages: Dict[str, str] = {}

        self.entryPoint: Optional[CodeType] = None

        self.addValidator(PythonFileValidator(self.ALLOWED_STRICT_MAIN_NAMES))
        self.addValidator(RequirementsValidator())
        self.addValidator(PackageValidator())

    def enableTestFiles(self: Builder, enableTestFiles: bool = True) -> Builder:
        self.testFilesEnabled = enableTestFiles
        return self

    def enableRequirements(self: Builder, enableRequirements: bool = True) -> Builder:
        self.requirementsEnabled = enableRequirements
        return self

    def enableLooseMainMatching(self: Builder, enableLooseMainMatching: bool = True) -> Builder:
        self.looseMainMatchingEnabled = enableLooseMainMatching
        return self

    def addPackage(self: Builder, packageName: str, packageVersion: Optional[str] = None) -> Builder:
        self.extraPackages[packageName] = packageVersion if packageVersion is not None else ""
        return self

    def addPackages(self: Builder, packages: List[Dict[str, str]]) -> Builder:
        for package in packages:
            self.extraPackages.update({package['name']: package['version']})

        return self

    def _addFileToMap(self, path: str, fileType: FileTypeMap) -> None:
        if fileType not in self.discoveredFileMap.keys():
            self.discoveredFileMap[fileType] = []

        self.discoveredFileMap[fileType].append(path)

    def _discoverSubmittedFiles(self, directoryToSearch: str) -> None:
        pathsToVisit: Iterable[str] = filter(filterSearchResults, os.listdir(directoryToSearch))

        if not pathsToVisit:
            return

        for path in pathsToVisit:
            if os.path.isdir(os.path.join(directoryToSearch, path)):
                self._discoverSubmittedFiles(os.path.join(directoryToSearch, path))
                continue

            if self.getTestFilesEnabled() and self.TEST_FILE_REGEX.match(path):
                self._addFileToMap(os.path.join(directoryToSearch, path), FileTypeMap.TEST_FILES)
                continue

            if self.PYTHON_FILE_REGEX.match(path):
                self._addFileToMap(os.path.join(directoryToSearch, path), FileTypeMap.PYTHON_FILES)
                continue

            if self.getRequirementsEnabled() and self.REQUIREMENTS_REGEX.match(path):
                self._addFileToMap(os.path.join(directoryToSearch, path), FileTypeMap.REQUIREMENTS)

    def _loadRequirements(self) -> None:
        if not self.getRequirementsEnabled() or FileTypeMap.REQUIREMENTS not in self.discoveredFileMap:
            return

        with open(self.discoveredFileMap[FileTypeMap.REQUIREMENTS][0], 'r') as r:
            # we are going to ignore any paths that aren't set up as package==version
            for line in r:
                line = line.strip()
                # we might want to add logging + telementry
                if not self.REQUIREMENTS_LINE_REGEX.match(line):
                    # we might want to add some logging here
                    continue
                line = line.split('==')

                self.addPackage(line[0], line[1] if len(line) == 2 else None)

    def _installRequirements(self) -> None:
        if not self.getRequirementsEnabled() or not self.extraPackages:
            return

        for package, version in self.extraPackages.items():
            try:
                subprocess.check_call([sys.executable, "-m", "pip", "install",
                                       f"{package}=={version}" if version else package],
                                      stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            # this isn't testable :( – basically this reattempts install if it fails by adding the magic flag
            except subprocess.CalledProcessError as _:  # pragma: no cover
                try:  # pragma: no cover
                    subprocess.check_call([sys.executable, "-m", "pip", "install",  # pragma: no cover
                                           f"{package}=={version}" if version else package, "--break-system-packages"], # pragma: no cover

                                          stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # pragma: no cover
                except subprocess.CalledProcessError as error:  # pragma: no cover
                    raise Exception(f"Failed to install '{package}'!")  # pragma: no cover

    def _identifyMainFile(self) -> str:
        if self.getLooseMainMatchingEnabled():
            return self.discoveredFileMap[FileTypeMap.PYTHON_FILES][0]

        for file in self.discoveredFileMap[FileTypeMap.PYTHON_FILES]:
            if os.path.basename(file) in self.ALLOWED_STRICT_MAIN_NAMES:
                return file

        # unreachable
        raise RuntimeError("Failed to identify main file! This error should be caught by a validator")

    def _readMainFile(self, mainPath) -> str:
        with open(mainPath, 'r', encoding="UTF-8") as r:
            return r.read()

    def _compileFile(self, filePath, code: str) -> CodeType:
        return compile(code, filePath, "exec")

    def doLoad(self):
        self._discoverSubmittedFiles(self.getSubmissionRoot())
        self._loadRequirements()

    def doBuild(self):
        self._installRequirements()
        mainFilePath = self._identifyMainFile()
        mainFileCode = self._readMainFile(mainFilePath)

        self.entryPoint = self._compileFile(mainFilePath, mainFileCode)

        # Huge todo here - This will be a seperate story i think
        # Basically by creating a meta hook in the import system, we can resolve modules from the students submission.
        # This also allows us to mock out imported libraries.
        # Im thinking this might be a seperate module as there is a decent amount of machinary that we need to override.

    def getExecutableSubmission(self) -> CodeType:
        if self.entryPoint is None:
            raise RuntimeError("Submission has not been built! No entrypoint has been defined!")
        return self.entryPoint

    def TEST_ONLY_removeRequirements(self):
        if not self.getRequirementsEnabled() or not self.extraPackages:
            return

        for package in self.extraPackages.keys():
            subprocess.check_call([sys.executable, "-m", "pip", "uninstall",
                                   "-y", package],
                                  stdout=subprocess.DEVNULL)

    def getTestFilesEnabled(self) -> bool:
        return self.testFilesEnabled

    def getRequirementsEnabled(self) -> bool:
        return self.requirementsEnabled

    def getLooseMainMatchingEnabled(self) -> bool:
        return self.looseMainMatchingEnabled

    def getDiscoveredFileMap(self) -> Dict[FileTypeMap, List[str]]:
        return self.discoveredFileMap

    def getExtraPackages(self) -> Dict[str, str]:
        return self.extraPackages



from importlib.abc import MetaPathFinder
from typing import List


class AbstractModuleFinder(MetaPathFinder):
    def __init__(self) -> None:
        super().__init__()
        self.modulesToReload: List[str] = []

    def getModulesToReload(self) -> List[str]:
        return self.modulesToReload



import abc
import argparse
import unittest.loader
from argparse import ArgumentParser
from typing import List, Callable, Dict, Optional
from unittest import TestSuite

import autograder_platform
from autograder_platform.config.Config import AutograderConfigurationBuilder, AutograderConfigurationProvider, \
    AutograderConfiguration

class AutograderCLITool(abc.ABC):

    PACKAGE_ERROR: str = "Required Package Error"
    SUBMISSION_ERROR: str = "Student Submission Error"
    ENVIRONMENT_ERROR: str = "Environment Error"
    RED_COLOR: str = u"\u001b[31m"
    YELLOW_COLOR: str = u"\u001b[33m"
    BLUE_COLOR: str = u"\u001b[34m"
    RESET_COLOR: str = u"\u001b[0m"

    @classmethod
    def print_error_message(cls, errorType: str, errorText: str) -> None:
        """
        This function prints out a validation error message as they occur.
        The error type is colored red when it is printed
        the format used is `[<error_type>] <error_text>`
        :param errorType: The error type
        :param errorText: the text for the error
        :return:
        """
        print(f"[{cls.RED_COLOR}{errorType}{cls.RESET_COLOR}]: {errorText}")


    @classmethod
    def print_warning_message(cls, warningType: str, warningText: str) -> None:
        print(f"[{cls.YELLOW_COLOR}{warningType}{cls.RESET_COLOR}]: {warningText}")

    @classmethod
    def print_info_message(cls, text: str) -> None:
        print(f"[{cls.BLUE_COLOR}INFORMATION{cls.RESET_COLOR}]: {text}")

    def __init__(self, tool_name: str):
        self.config: Optional[AutograderConfiguration] = None
        self.arguments: Optional[argparse.Namespace] = None
        self.tests: Optional[TestSuite] = None

        self.parser: ArgumentParser = argparse.ArgumentParser(description=f"Autograder Platform - {tool_name}")

        # required CLI arguments
        self.parser.add_argument("--config-file", default="./config.toml",
                            help="Set the location of the config file")

    @staticmethod
    def get_version() -> str:
        return autograder_platform.__version__

    @abc.abstractmethod
    def configure_options(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def set_config_arguments(self, configBuilder: AutograderConfigurationBuilder[AutograderConfiguration]):
        raise NotImplementedError()

    @abc.abstractmethod
    def run(self) -> bool:
        raise NotImplementedError()

    def load_config(self):  # pragma: no cover
        self.arguments = self.parser.parse_args()

        # load toml then override any options in toml with things that are passed to the runtime
        builder = AutograderConfigurationBuilder() \
            .fromTOML(file=self.arguments.config_file)

        self.set_config_arguments(builder)

        self.config = builder.build()

        AutograderConfigurationProvider.set(self.config)

    def discover_tests(self):  # pragma: no cover
        self.tests = unittest.loader.defaultTestLoader.discover(self.config.config.test_directory)




from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from types import ModuleType
from typing import Optional, List, Dict

from autograder_platform.StudentSubmissionImpl.Python.AbstractPythonImportFactory import AbstractModuleFinder
from autograder_platform.TestingFramework.SingleFunctionMock import SingleFunctionMock


class MockedModuleFinder(AbstractModuleFinder, Loader):
    def __init__(self, name: str, module: ModuleType, mocks: Dict[str, SingleFunctionMock]) -> None:
        self.name: str = name
        self.module: ModuleType = module
        self.mocks: Dict[str, SingleFunctionMock] = mocks
        self.modulesToReload: List[str] = [self.name]

    def create_module(self, spec: ModuleSpec) -> ModuleType:
        return self.module

    def exec_module(self, module):
        for methodName, mock in self.mocks.items():
            splitName = methodName.split('.')

            if mock.spy:
                mock.setSpyFunction(getattr(module, splitName[-1]))

            setattr(module, splitName[-1], mock)

    def find_spec(self, fullname, path, target=None) -> Optional[ModuleSpec]:
        if self.name != fullname:
            return None

        return ModuleSpec(self.name, self)

    def getModulesToReload(self) -> List[str]:
        return self.modulesToReload




from .common import ValidationHook
from .AbstractValidator import AbstractValidator
from os import PathLike
import os
from typing import Union

class SubmissionPathValidator(AbstractValidator):
    @staticmethod
    def getValidationHook() -> ValidationHook:
        return ValidationHook.PRE_LOAD

    def __init__(self):
        super().__init__()
        self.pathToValidate: Union[PathLike, str] = ""

    def setup(self, studentSubmission):
        self.pathToValidate = studentSubmission.getSubmissionRoot()
        
    def run(self):
        if not os.path.exists(self.pathToValidate):
            self.addError(
                FileNotFoundError(f"{self.pathToValidate} does not exist or is not accessible!")
            )
            return

        if not os.path.isdir(self.pathToValidate):
            self.addError(
                NotADirectoryError(f"{self.pathToValidate} is not a directory!")
            )

        if not os.access(self.pathToValidate, os.R_OK):
            self.addError(
                PermissionError(f"Unable to read from {self.pathToValidate}!")
            )

        if len(os.listdir(self.pathToValidate)) < 1:
            self.addError(
                FileNotFoundError(
                    f"No files found in {self.pathToValidate}.\n"
                    f"Ensure that your submission is in {self.pathToValidate}."
                )
            )

